#!/usr/bin/perl

# Name: IAM, Interactive Account Manager
# Description: Multi server Multi domain account manager
# Version: 0.8
# Authors: Jason Jorgensen <jasonj@innominatus.com>
# Copyright (C) 2003  Jason Jorgensen
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
###########################################################################


push @INC, "/usr/local/iam";

use strict;
use IO::Socket;
use IO::Socket::SSL;
use POSIX 'WNOHANG';
use POSIX 'strftime';
use Authen::Libwrap qw( hosts_ctl STRING_UNKNOWN );
use Sys::Syslog;
use MD5;
use Crypt::Random qw( makerandom );
use IAM;
use File::stat;


### Modules that I have tried but didnt give the results I wanted
#use String::Random;
#use Data::Random qw(:all);
#use Crypt::HCE_MD5;
#use Crypt::PasswdMD5;
#use Crypt::OTP;

## This stuff to be moved to a config file
#my $listen_interface = "localhost";
my $listen_interface = "0.0.0.0";
my $listen_port = "1077";
my $plugin_dir = "/etc/IAM/plugins";
my $comm_method = "ssl";         # Use SSL or some other method for the communications
my $masterslave = "master";        # master or slave 
my $driver = "mysql";
my $masterdatabase = "IAM_MASTER";
my $remotedatabase = "REMOTES";
my $hostname = "127.0.0.1";
my $port = "3306";
my $user = "iam";
my $password = "iammasterpass";
my $mastermodule = "IAM";
my $iam; # IAM main object
##

my @months = ("January","February","March","April","May","June",
           "July","August","September","October","November","December");

my ($dev,$ino,$mode,$nlink,$uid,$gid,$rdev,$size,$atime,$mtime,$ctime,$blksize,$blocks) = stat("$0");
my ($sec,$min,$hour,$mday,$mon,$year,$wday,$yday,$isdst) = localtime($mtime); # or gmtime($mtime);
my $time = sprintf("%02d",$min);
my $year += 1900;
            
my $scriptlastmodified = "$months[$mon] $mday, $year at $hour:$time";
my $scriptversion = "0.8";
my $scriptname = "IAMd";
my ($physcriptname) = ( $0 =~ /.*\/(.*)/ );
my $configdir = "/etc/IAM";
my $plugins;
my $clientaddress;
my $daemonmethod = "fork";

my $debug = 0;
my $quit = 0;
my $childcounter;


die "Exiting... No server pad: $! \n" unless my $padding = &loadpadding();  # get the secret pass phrase

#Handle signals Start
$SIG{CHLD} = sub {while ( waitpid (-1, WNOHANG) >0 ) { } };
$SIG{INT} = $SIG{TERM} = sub { $quit++ };
$SIG{HUP} = sub { exec ($0, @ARGV) or die "Exiting... Couldnt restart. Going down in flames! : $! \n"; };
#Handle signals End

&handle_args(); # Handle command line arguments.

my @plugin_names = &plugin_names();  # get the names of the available plugins

my $listen_socket = &open_for_business($comm_method, $listen_port, $listen_interface);
print "$scriptname ready for connections...\n";


if ($daemonmethod eq "fork") {
  $childcounter = 0;
  while ( ! $quit ) {    # The main loop. What makes this a server.
    next unless my $connection = $listen_socket->accept();
    $childcounter++;
    print "DEBUG: COUNTER: $childcounter\n" if $debug;
    defined(my $child = fork()) or die "Exiting... Couldnt fork: $!\n";
    if ($child == 0) {   ## I am the child, do some stuff.
      die "Exiting... error with connection: $!\n" unless $connection;
    
      my $wrapcheckresult = &tcpwrapme($connection);               # TCP Wrapper the connection
      my $otpcheckresult = &authenticateconnection($connection);   # Authenticate the connector
    
      $listen_socket->close;
      if ($wrapcheckresult & $otpcheckresult) { 
        print $connection "Welcome\n";
	&iam_connect() if $masterslave eq "master";
        interact($connection);      # interact with the client. answer the clients requests
	&iam_disconnect() if $masterslave eq "master";
      }
      else { print $connection "Failure!\n"; }
      $connection->close;
      exit 0;  # child die! you are no longer required
    }
    $connection->close;  # parent closing connection so child can handle the business
  }
}
elsif ($daemonmethod eq "select") {
  use IO::Select;
  my @connections;
  my $connectionmanager = new IO::Select($listen_socket) || die "Cant open socket manager IO::Select: $@\n";
  &iam_connect() if $masterslave eq "master";
  while ( (@connections = $connectionmanager->can_read) && (! $quit )) {    # The main loop. What makes this a server.
    foreach my $connection (@connections) {
      if ($connection == $listen_socket) {
        #create new socket
        print "DEBUG:IAMd: new socket connection, creating new handle\n";
        my $newconnection = $listen_socket->accept();
        my $wrapcheckresult = &tcpwrapme($newconnection);               # TCP Wrapper the connection
        my $otpcheckresult = &authenticateconnection($newconnection);   # Authenticate the connector
        if ($wrapcheckresult & $otpcheckresult) { 
          print $newconnection "Welcome\n";
          $connectionmanager->add($newconnection);
        }
        else { print $connection "Failure!\n"; }

      }
      else {
        # Process existing socket
        print "DEBUG:IAMd: existing socket, do some work.\n";
        &interact($connection);      # interact with the client. answer the clients requests
        $connectionmanager->remove($connection);
        close($connection);
      }
    }
  }
  &iam_disconnect() if $masterslave eq "master";
  
}
else { die "Invalid daemon method selected\n"; }
exit;

## Daemon subroutines
sub authenticateconnection {
  my $connection = shift;
  my $OTP;
  my $OTPelement;
  my $OTPclientchallenge;
  my $OTPclientresponse;
  my $otpcheckresult;

    $OTPclientchallenge = makerandom ( Size => 64, Strength => 0 );
    $OTPelement = new MD5;
    $OTPelement->add($padding);
    $OTPelement->add($OTPclientchallenge);
    $OTP = unpack("H*", $OTPelement->digest());
    print "DEBUG:IAMd: Generated OTP = $OTP\n" if $debug;
    print "DEBUG:IAMd: Challenging Client with OTP = $OTPclientchallenge\n" if $debug;
    print $connection "$OTPclientchallenge\n";
    $OTPclientresponse = <$connection>;
    chomp $OTPclientresponse;
    print "DEBUG:IAMd: Client said = $OTPclientresponse\n" if $debug;
    if ($OTP eq $OTPclientresponse) { print "DEBUG:IAMd: Client responded correctly\n" if $debug; $otpcheckresult = 1; }
 
  return $otpcheckresult;
}

sub clientrequests {
  my $socket = shift;
  my $request = shift;
  
  print "DEBUG:IAMd: sub clientrequests being called with '$request'\n" if $debug;
  if ($request =~ /^\s*$/) { print "DEBUG:IAMd: erroneous request\n" if $debug; print $socket "1 no request sent\n"; return;}
  elsif ($request =~ /^exit$|^quit$|^logout$|^bye$/i ) { print $socket "GoodBye\n"; }
  elsif ($request eq "list_plugins" ) { print $socket &request_list_plugins()."\n"; }
  elsif ($request eq "version" ) { print $socket &request_version(); }
  elsif ($request =~ /^update\_currentuser.*$/) { print "DEBUG:IAMd: calling update_currentuser with: $request\:\:$clientaddress\n" if $debug; print $socket &iam_request($request.'::'.$clientaddress); }
  elsif ($request =~ /^remote\_.*/i) { 
    print "DEBUG:IAMd: Account request\n" if $debug;
    $request =~ s/^remote_//i;
    my $result = &remote_request($request);
    #print $socket &remote_request($request);
    print "DEBUG:IAMd:  result: $result\n" if $debug;
    print "DEBUG:IAMd: Child Counter: $childcounter\n" if $debug;
    if ($result !~ /\n$/) { $result .= "\n"; }
    print $socket "$result";
  }
  elsif ($request !~ /^\s+$/) { 
    print "DEBUG:IAMd: User request\n" if $debug;
    my $result = &iam_request($request);
    print "DEBUG:IAMd: result: $result\n" if $debug;
#--    if ($result !~ /\n$/) { $result .= "\n"; }
    print $socket $result;
  }
  else { 
    print "DEBUG:IAMd: Unknown request\n" if $debug;
    print $socket "1 Unknown request\n"; 
    return;
  }
  print $socket "_ENDRESULT_\n"; ## a newline to signal end of request
  print "DEBUG:IAMd: sending end of result code to client\n" if $debug; ## a newline to signal end of request
  return 1;
}

sub iam_connect {
  $iam = eval {IAM->new($plugindir)};   ## establish IAM object
  $iam->connect($driver, $masterdatabase, $remotedatabase, $hostname, $port, $user, $password);   ## establish connection to IAM databases
  die "Exiting... Cant open master database\n" unless $iam;
  $iam->debug_on if $debug;
}

sub iam_disconnect {
  $iam->disconnect;
  $iam = undef;
}

sub interact {
  my $socket = shift;

  my $clientcommunication;
  
  while ($socket->connected && $clientcommunication !~ /^exit$|^quit$|^logout$|^bye$/i) {                  # Loop until the client chooses to disconnect
    #print $socket "> ";  ## prompt
    print "DEBUG:IAMd: waiting for client request.\n" if $debug;
    $clientcommunication = <$socket>;
    if ($clientcommunication eq '') {
      print "DEBUG:IAMd: client is no longer connected\n" if $debug;
      $socket->shutdown(2);
      return;
    }
    chomp $clientcommunication;
    # handle client requests
    &clientrequests($socket, $clientcommunication);
  }
}

sub handle_args {
  my $arg;

  foreach $arg (@ARGV) {
    if ($arg =~ /^\-\-daemon$|^\-d$/) {
      my $pid = fork;
      exit if $pid;
      die "Exiting... Couldnt enter daemon mode. : $! \n" unless defined($pid);
    }
    if ($arg =~ /^\-\-version$|^\-v$/) {
      print &version_info();
      exit 0;
    }
    if ($arg =~ /^\-\-debug$|^\-D$/) {
      $debug = 1;
    }
    if ($arg =~ /^\-\-fork$|^\-F$/) {
      $daemonmethod = "fork";
    }
    if ($arg =~ /^\-\-multisocket$|^\-M$/) {
      $daemonmethod = "select";
    }
    if ($arg =~ /^\-\-thread$|^\-T$/) {
      $daemonmethod = "thread";
    }
    if ($arg =~ /^\-\-help$|^\-h$/) {
      print "$scriptname Version: $scriptversion Last Modified: $scriptlastmodified\n";
      print "  Usage: $physcriptname (options)\n";
      print "    -d   --daemon      Puts $scriptname into DAEMON mode and launchs into the background\n";
      print "    -D   --debug       Places $scriptname into DEBUG and displays alot more messages\n";
      print "    -h   --help        Displays this help info and exits\n";
      print "    -v   --version     Displays version information and exits\n";
      print "\n";
      print "    -F   --fork        Daemon will fork to handle requests (default)\n";
      print "                         (Good if you have accounts on the machine with $scriptname)\n";
      print "    -M   --multisocket Daemon will handle many socket connections \n";
      print "                         (Good for speed if you dont have accounts on the machine with $scriptname)\n";
      print "    -T   --thread      (Not implemented yet)\n";
      exit 0;
    }
  }
}

sub loadpadding {
  open (PADDING, "$configdir/server_pad");
  my $padding = <PADDING>;
  close (PADDING);
  wait;

  return $padding;
}

sub plugin_names {
  return "personalwebspace";
}


sub open_for_business {
  my $method = shift;
  my $port = shift;
  my $interface = shift;
  my $socket;

  print "DEBUG:IAMd: sub open_for_business called\n" if $debug;
  print "DEBUG:IAMd: Interface: $interface | Port: $port | Method: $method\n" if $debug;
  if ($method eq "ssl") {
    #$IO::Socket::SSL::DEBUG=1;
    $socket = IO::Socket::SSL->new(LocalAddr => $interface,
                                   LocalPort => $port,
                                   Proto     => 'tcp',
                                   Reuse     => 1,
                                   Timeout   => 4,
                                   Listen    => 10,
                                   SSL_use_cert    => 0,
                                   SSL_server      => 1,
                                   SSL_verify_mode => 0x00,
                                   SSL_key_file    => "$configdir/certs/IAMd-server.key",
                                   SSL_cert_file   => "$configdir/certs/IAMd-server.crt",
                                   );
                                   #SSL_ca_path     => '$configdir/certs',
                                   #SSL_ca_file     => '$configdir/certs/private.crs',
                                   #);
  }
  elsif ($method eq "plaintext") {
    $socket = IO::Socket::INET->new(LocalAddr => $interface,
                                    LocalPort => $port,
                                    Proto     => 'tcp',
                                    Reuse     => 1,
                                    Timeout   => 4,
                                    Listen    => 10,
                                    );
  }
  else { die "Unknow method\n"; }
  if ($socket) { $socket->autoflush(1); return $socket; }
  else { die "Exiting... Couldnt open communications on interface '$interface' port '$port' with method '$method': $@ \n"; }
}

sub tcpwrapme {
  my $connection = shift;

  my $clientaddresspacked = $connection->peeraddr();
  $clientaddress = inet_ntoa($clientaddresspacked);
  print "DEBUG:IAMd: Connecting peer from: $clientaddress\n" if $debug;
  my $clientfqdn = gethostbyaddr($clientaddresspacked, AF_INET);
  print "DEBUG:IAMd: Connecting peer from: $clientfqdn\n" if $debug;
  my $wrapcheckresult = hosts_ctl("$scriptname",
                               "$clientaddress",
                               "$clientfqdn",
                               STRING_UNKNOWN );
  print "DEBUG:IAMd: Access is ", $wrapcheckresult ? "granted" : "refused", "\n" if $debug;

  return $wrapcheckresult;
}

sub version_info {
  return "$scriptname Version: $scriptversion Last Modified: $scriptlastmodified\n"; 
}


## User request subroutines
sub request_list_plugins {

  print "DEBUG:IAMd: sub list_plugins being called\n" if $debug;
  return "@plugin_names";
}

sub request_version {
  return &version_info();
}

## Database interaction subroutines
sub remote_request {
  my $request = shift @_;

  my $endresult;

  (my $function, my $requiredplugin, my @args) = split /::/, $request;

  if (! $function) { return "1 No IAM 'function' specified"; }
  if (! $requiredplugin) { return "1 No 'pluginname' specified"; }

  print "DEBUG:IAMd: $requiredplugin\n" if $debug;
  my $usestatement = "use $requiredplugin"; # setup the 'use' statement to be eval'd
  
  eval $usestatement;
  print "DEBUG:IAMd: eval'ing $usestatement: $@\n" if $debug;
  my $iamplugin = eval { $requiredplugin->new(); };  # calling new method on the dynamicly loaded object "plugin"
  
  *fullfunction = $requiredplugin."::".$function;
  print "DEBUG:IAMd: exists?: $function\n" if $debug;
  if (defined &fullfunction) {                  # handle object methods
    print "DEBUG:IAMd: calling IAM plugin with function: $function: @args\n" if $debug;
    my $result = $iamplugin->$function(@args);                      # call plugin method
    $endresult .= $result;
    chomp $result if $debug;
    print "DEBUG:IAMd: request response: $result\n" if $debug;
  }
  elsif ($iamplugin->{$function}) {                   # handle object variables
    print "DEBUG:IAMd: calling IAM plugin with variable: $function\n" if $debug;
    my $result = $iamplugin->{$function};
    $endresult .=  $result;
    chomp $result if $debug;
    print "DEBUG:IAMd: request response: $result\n" if $debug;
  }
  else {
    print "DEBUG:IAMd: Unknown command requested: $request\n" if $debug;
    $endresult .= "11 Unknown IAM Command\n";
  }
  if ($endresult !~ /.*\n$/) { $endresult .= "\n"; }  # if it doesnt end in a newline, add one so the client sees the info

  $iamplugin->close;
  $iamplugin=undef;
  return $endresult;
}

sub iam_request {
  my $request = shift @_;
  
  my $endresult;
  (my $function, my @args) = split /::/, $request;
  *fullfunction = $mastermodule."::".$function;
  print "DEBUG:IAMd: exists?: $function\n" if $debug;
  if (defined &fullfunction) {                  # handle object methods
    print "DEBUG:IAMd: calling IAM with function: $function: @args\n" if $debug;
    my $result = $iam->$function(@args);
    $endresult .= $result;
    chomp $result if $debug;
    print "DEBUG:IAMd: request response: $result\n" if $debug;
  }
  elsif ($iam->{$function}) {                   # handle object variables
    print "DEBUG:IAMd: calling IAM with variable: $function\n" if $debug;
    my $result = $iam->{$function};
    $endresult .=  $result;
    chomp $result if $debug;
    print "DEBUG:IAMd: request response: $result\n" if $debug;
  }
  else {
    print "DEBUG:IAMd: Unknown command requested: $request\n" if $debug;
    $endresult .= "1 Unknown Command\n";
  }
  if ($endresult !~ /\n$/) { $endresult .= "\n"; }  # if it doesnt end in a newline, add one so the client sees the info

  print "DEBUG:IAMd: endrequest: $endresult\n" if $debug;
  print "DEBUG:IAMd: Counter: $childcounter\n" if $debug;
  return $endresult;
}
